/*
 * Copyright (c) 2017 SpaceToad and the BuildCraft team
 * This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0. If a copy of the MPL was not
 * distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/
 */

package buildcraft.builders.snapshot;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.google.common.base.Predicates;

import org.lwjgl.opengl.GL11;
import org.lwjgl.util.glu.GLU;

import net.minecraft.client.Minecraft;
import net.minecraft.client.gui.ScaledResolution;
import net.minecraft.client.renderer.BufferBuilder;
import net.minecraft.client.renderer.GlStateManager;
import net.minecraft.client.renderer.WorldVertexBufferUploader;
import net.minecraft.client.renderer.texture.TextureMap;
import net.minecraft.client.renderer.tileentity.TileEntityRendererDispatcher;
import net.minecraft.client.renderer.vertex.DefaultVertexFormats;
import net.minecraft.entity.Entity;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.Vec3d;

import net.minecraftforge.fml.relauncher.Side;
import net.minecraftforge.fml.relauncher.SideOnly;

import buildcraft.lib.net.MessageManager;

public enum ClientSnapshots {
    INSTANCE;

    private final List<Snapshot> snapshots = new ArrayList<>();
    private final List<Snapshot.Key> pending = new ArrayList<>();
    private final Map<Snapshot.Key, FakeWorld> worlds = new HashMap<>();
    private final Map<Snapshot.Key, BufferBuilder> buffers = new HashMap<>();

    public Snapshot getSnapshot(Snapshot.Key key) {
        Snapshot found = snapshots.stream().filter(snapshot -> snapshot.key.equals(key)).findFirst().orElse(null);
        if (found == null && !pending.contains(key)) {
            pending.add(key);
            MessageManager.sendToServer(new MessageSnapshotRequest(key));
        }
        return found;
    }

    public void onSnapshotReceived(Snapshot snapshot) {
        pending.remove(snapshot.key);
        snapshots.add(snapshot);
    }

    @SideOnly(Side.CLIENT)
    public void renderSnapshot(Snapshot.Header header, int offsetX, int offsetY, int sizeX, int sizeY) {
        if (header == null) {
            return;
        }
        Snapshot snapshot = getSnapshot(header.key);
        if (snapshot == null) {
            return;
        }
        renderSnapshot(snapshot, offsetX, offsetY, sizeX, sizeY);
    }

    @SideOnly(Side.CLIENT)
    public void renderSnapshot(Snapshot snapshot, int offsetX, int offsetY, int sizeX, int sizeY) {
        FakeWorld world = worlds.computeIfAbsent(snapshot.key, key -> {
            FakeWorld localWorld = new FakeWorld();
            localWorld.uploadSnapshot(snapshot);
            return localWorld;
        });
        BufferBuilder bufferBuilder = buffers.computeIfAbsent(snapshot.key, key -> {
            BufferBuilder localBuffer = new BufferBuilder(1024) {
                @Override
                public void func_178965_a() {
                }
            };
            localBuffer.func_181668_a(GL11.GL_QUADS, DefaultVertexFormats.field_176600_a);
            for (int z = 0; z < snapshot.size.func_177952_p(); z++) {
                for (int y = 0; y < snapshot.size.func_177956_o(); y++) {
                    for (int x = 0; x < snapshot.size.func_177958_n(); x++) {
                        BlockPos pos = new BlockPos(x, y, z).func_177971_a(FakeWorld.BLUEPRINT_OFFSET);
                        localBuffer.func_178969_c(
                            -FakeWorld.BLUEPRINT_OFFSET.func_177958_n(),
                            -FakeWorld.BLUEPRINT_OFFSET.func_177956_o(),
                            -FakeWorld.BLUEPRINT_OFFSET.func_177952_p()
                        );
                        Minecraft.func_71410_x().func_175602_ab().func_175018_a(
                            world.func_180495_p(pos),
                            pos,
                            world,
                            localBuffer
                        );
                        localBuffer.func_178969_c(0, 0, 0);
                    }
                }
            }
            localBuffer.func_178977_d();
            return localBuffer;
        });
        GlStateManager.func_179123_a();
        GlStateManager.func_179126_j();
        GlStateManager.func_179147_l();
        GlStateManager.func_179094_E();
        GlStateManager.func_179128_n(GL11.GL_PROJECTION);
        GlStateManager.func_179094_E();
        GlStateManager.func_179096_D();
        ScaledResolution scaledResolution = new ScaledResolution(Minecraft.func_71410_x());
        int viewportX = offsetX * scaledResolution.func_78325_e();
        int viewportY = Minecraft.func_71410_x().field_71440_d - (sizeY + offsetY) * scaledResolution.func_78325_e();
        int viewportWidth = sizeX * scaledResolution.func_78325_e();
        int viewportHeight = sizeY * scaledResolution.func_78325_e();
        GL11.glEnable(GL11.GL_SCISSOR_TEST);
        GL11.glScissor(
            viewportX,
            viewportY,
            viewportWidth,
            viewportHeight
        );
        GlStateManager.func_179086_m(GL11.GL_DEPTH_BUFFER_BIT);
        GL11.glDisable(GL11.GL_SCISSOR_TEST);
        GlStateManager.func_179083_b(
            viewportX,
            viewportY,
            viewportWidth,
            viewportHeight
        );
        GlStateManager.func_179152_a(scaledResolution.func_78325_e(), scaledResolution.func_78325_e(), 1);
        GLU.gluPerspective(70.0F, (float) sizeX / sizeY, 0.1F, 1000.0F);
        GlStateManager.func_179128_n(GL11.GL_MODELVIEW);
        GlStateManager.func_179096_D();
        GlStateManager.func_179091_B();
        GlStateManager.func_179094_E();
        int snapshotSize = Math.max(Math.max(snapshot.size.func_177958_n(), snapshot.size.func_177956_o()), snapshot.size.func_177956_o());
        GlStateManager.func_179109_b(0, 0, -snapshotSize * 2F - 3);
        GlStateManager.func_179114_b(20, 1, 0, 0);
        GlStateManager.func_179114_b((System.currentTimeMillis() % 3600) / 10F, 0, 1, 0);
        GlStateManager.func_179109_b(-snapshot.size.func_177958_n() / 2F, -snapshot.size.func_177956_o() / 2F, -snapshot.size.func_177952_p() / 2F);
        GlStateManager.func_179109_b(0, snapshotSize * 0.1F, 0);
        Minecraft.func_71410_x().func_175598_ae().field_78724_e.func_110577_a(TextureMap.field_110575_b);
        new WorldVertexBufferUploader().func_181679_a(bufferBuilder);
        if (snapshotSize < 32) {
            TileEntityRendererDispatcher.field_147556_a.preDrawBatch();
            for (int z = 0; z < snapshot.size.func_177952_p(); z++) {
                for (int y = 0; y < snapshot.size.func_177956_o(); y++) {
                    for (int x = 0; x < snapshot.size.func_177958_n(); x++) {
                        BlockPos pos = new BlockPos(x, y, z).func_177971_a(FakeWorld.BLUEPRINT_OFFSET);
                        GlStateManager.func_179123_a();
                        // noinspection ConstantConditions
                        TileEntityRendererDispatcher.field_147556_a.func_147549_a(
                            world.func_175625_s(pos),
                            pos.func_177958_n() - FakeWorld.BLUEPRINT_OFFSET.func_177958_n(),
                            pos.func_177956_o() - FakeWorld.BLUEPRINT_OFFSET.func_177956_o(),
                            pos.func_177952_p() - FakeWorld.BLUEPRINT_OFFSET.func_177952_p(),
                            0
                        );
                        GlStateManager.func_179099_b();
                    }
                }
            }
            TileEntityRendererDispatcher.field_147556_a.drawBatch(1);
        }
        // noinspection Guava
        for (Entity entity : world.func_175644_a(Entity.class, Predicates.alwaysTrue())) {
            Vec3d pos = entity.func_174791_d();
            GlStateManager.func_179123_a();
            Minecraft.func_71410_x().func_175598_ae().func_188391_a(
                entity,
                pos.field_72450_a - FakeWorld.BLUEPRINT_OFFSET.func_177958_n(),
                pos.field_72448_b - FakeWorld.BLUEPRINT_OFFSET.func_177956_o(),
                pos.field_72449_c - FakeWorld.BLUEPRINT_OFFSET.func_177952_p(),
                0,
                0,
                true
            );
            GlStateManager.func_179099_b();
        }
        GlStateManager.func_179121_F();
        GlStateManager.func_179101_C();
        GlStateManager.func_179128_n(GL11.GL_PROJECTION);
        GlStateManager.func_179083_b(0, 0, Minecraft.func_71410_x().field_71443_c, Minecraft.func_71410_x().field_71440_d);
        GlStateManager.func_179121_F();
        GlStateManager.func_179128_n(GL11.GL_MODELVIEW);
        GlStateManager.func_179121_F();
        GlStateManager.func_179084_k();
        GlStateManager.func_179097_i();
        GlStateManager.func_179099_b();
    }
}
