/* Copyright (c) 2016 SpaceToad and the BuildCraft team
 *
 * This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0. If a copy of the MPL was not
 * distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/. */
package buildcraft.robotics.gui;

import java.io.IOException;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.vecmath.Vector3d;

import org.apache.commons.lang3.tuple.Pair;
import org.lwjgl.BufferUtils;
import org.lwjgl.input.Mouse;
import org.lwjgl.opengl.GL11;
import org.lwjgl.util.glu.GLU;

import net.minecraft.client.gui.ScaledResolution;
import net.minecraft.client.renderer.BufferBuilder;
import net.minecraft.client.renderer.GlStateManager;
import net.minecraft.client.renderer.RenderHelper;
import net.minecraft.client.renderer.Tessellator;
import net.minecraft.client.renderer.vertex.DefaultVertexFormats;
import net.minecraft.item.EnumDyeColor;
import net.minecraft.item.ItemStack;
import net.minecraft.util.ResourceLocation;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.ChunkPos;

import buildcraft.lib.gui.GuiBC8;
import buildcraft.lib.gui.GuiIcon;
import buildcraft.lib.gui.pos.GuiRectangle;

import buildcraft.core.BCCoreItems;
import buildcraft.core.item.ItemPaintbrush_BC8;
import buildcraft.robotics.container.ContainerZonePlanner;
import buildcraft.robotics.zone.ZonePlan;
import buildcraft.robotics.zone.ZonePlannerMapChunk;
import buildcraft.robotics.zone.ZonePlannerMapChunk.MapColourData;
import buildcraft.robotics.zone.ZonePlannerMapChunkKey;
import buildcraft.robotics.zone.ZonePlannerMapDataClient;
import buildcraft.robotics.zone.ZonePlannerMapRenderer;

public class GuiZonePlanner extends GuiBC8<ContainerZonePlanner> {
    private static final ResourceLocation TEXTURE_BASE = new ResourceLocation("buildcraftrobotics:textures/gui/zone_planner.png");
    private static final int SIZE_X = 256, SIZE_Y = 228;
    private static final GuiIcon ICON_GUI = new GuiIcon(TEXTURE_BASE, 0, 0, SIZE_X, SIZE_Y);
    private static final GuiIcon ICON_PROGRESS_INPUT = new GuiIcon(TEXTURE_BASE, 9, 228, 28, 9);
    private static final GuiIcon ICON_PROGRESS_OUTPUT = new GuiIcon(TEXTURE_BASE, 0, 228, 9, 28);
    private static final GuiRectangle RECT_PROGRESS_INPUT = new GuiRectangle(44, 128, 28, 9);
    private static final GuiRectangle RECT_PROGRESS_OUTPUT = new GuiRectangle(236, 45, 9, 28);
    private float startMouseX = 0;
    private float startMouseY = 0;
    private float startPositionX = 0;
    private float startPositionZ = 0;
    private float camY = 256;
    private float scaleSpeed = 0;
    private float positionX;
    private float positionZ;
    private boolean canDrag = false;
    private BlockPos lastSelected = null;
    private BlockPos selectionStartXZ = null;
    private ZonePlan bufferLayer = null;

    public GuiZonePlanner(ContainerZonePlanner container) {
        super(container);
        field_146999_f = SIZE_X;
        field_147000_g = SIZE_Y;
        BlockPos tilePos = container.tile.func_174877_v();
        positionX = tilePos.func_177958_n();
        positionZ = tilePos.func_177952_p();
    }

    private ItemStack getCurrentStack() {
        return field_146297_k.field_71439_g.field_71071_by.func_70445_o();
    }

    private ItemStack getPaintbrush() {
        ItemStack currentStack = getCurrentStack();
        if (!currentStack.func_190926_b() && currentStack.func_77973_b() instanceof ItemPaintbrush_BC8) {
            return currentStack;
        }
        return null;
    }

    private ItemPaintbrush_BC8.Brush getPaintbrushBrush() {
        ItemStack paintbrush = getPaintbrush();
        if (paintbrush != null) {
            ItemPaintbrush_BC8.Brush brush = BCCoreItems.paintbrush.getBrushFromStack(paintbrush);
            if (brush.colour != null) {
                return brush;
            }
        }
        return null;
    }

    @Override
    public void func_146274_d() throws IOException {
        int wheel = Mouse.getEventDWheel();
        if (wheel != 0) {
            scaleSpeed -= wheel / 30F;
        }
        super.func_146274_d();
    }

    @Override
    protected void func_73864_a(int mouseX, int mouseY, int mouseButton) throws IOException {
        super.func_73864_a(mouseX, mouseY, mouseButton);
        canDrag = false;
        if (getPaintbrush() != null) {
            if (lastSelected != null) {
                selectionStartXZ = lastSelected;
            }
        } else if (getCurrentStack().func_190926_b()) {
            startPositionX = positionX;
            startPositionZ = positionZ;
            startMouseX = mouseX;
            startMouseY = mouseY;
            canDrag = true;
        }
    }

    @Override
    protected void func_146273_a(int mouseX, int mouseY, int clickedMouseButton, long timeSinceLastClick) {
        super.func_146273_a(mouseX, mouseY, clickedMouseButton, timeSinceLastClick);
        if (!canDrag) {
            if (lastSelected != null && getPaintbrushBrush() != null) {
                bufferLayer = new ZonePlan(container.tile.layers[getPaintbrushBrush().colour.func_176765_a()]);
                if (selectionStartXZ != null && getPaintbrushBrush() != null && lastSelected != null) {
                    for (int x = Math.min(selectionStartXZ.func_177958_n(), lastSelected.func_177958_n());
                         x < Math.max(selectionStartXZ.func_177958_n(), lastSelected.func_177958_n());
                         x++) {
                        for (int z = Math.min(selectionStartXZ.func_177952_p(), lastSelected.func_177952_p());
                             z < Math.max(selectionStartXZ.func_177952_p(), lastSelected.func_177952_p());
                             z++) {
                            if (clickedMouseButton == 0) {
                                bufferLayer.set(
                                    x - container.tile.func_174877_v().func_177958_n(),
                                    z - container.tile.func_174877_v().func_177952_p(),
                                    true
                                );
                            } else if (clickedMouseButton == 1) {
                                bufferLayer.set(
                                    x - container.tile.func_174877_v().func_177958_n(),
                                    z - container.tile.func_174877_v().func_177952_p(),
                                    false
                                );
                            }
                        }
                    }
                }
            }
            return;
        }
        float deltaX = mouseX - startMouseX;
        float deltaY = mouseY - startMouseY;
        float s = 0.3F;
        positionX = startPositionX - deltaX * s;
        positionZ = startPositionZ - deltaY * s;
    }

    @Override
    protected void func_146286_b(int mouseX, int mouseY, int state) {
        super.func_146286_b(mouseX, mouseY, state);
        selectionStartXZ = null;
        if (getPaintbrushBrush() != null && bufferLayer != null) {
            container.tile.layers[getPaintbrushBrush().colour.func_176765_a()] = bufferLayer;
            container.tile.sendLayerToServer(getPaintbrushBrush().colour.func_176765_a());
        }
        bufferLayer = null;
    }

    @Override
    protected void drawBackgroundLayer(float partialTicks) {
        ICON_GUI.drawAt(mainGui.rootElement);

        drawProgress(
            RECT_PROGRESS_INPUT,
            ICON_PROGRESS_INPUT,
            container.tile.deltaProgressInput.getDynamic(partialTicks),
            1
        );
        drawProgress(
            RECT_PROGRESS_OUTPUT,
            ICON_PROGRESS_OUTPUT,
            1,
            container.tile.deltaProgressOutput.getDynamic(partialTicks)
        );
    }

    private BlockPos rayTrace(int screenX, int screenY) {
        BlockPos found = null;
        FloatBuffer projectionBuffer = BufferUtils.createFloatBuffer(16);
        FloatBuffer modelViewBuffer = BufferUtils.createFloatBuffer(16);
        IntBuffer viewportBuffer = BufferUtils.createIntBuffer(16);

        GlStateManager.func_179111_a(GL11.GL_PROJECTION_MATRIX, projectionBuffer);
        GlStateManager.func_179111_a(GL11.GL_MODELVIEW_MATRIX, modelViewBuffer);
        GlStateManager.func_187445_a(GL11.GL_VIEWPORT, viewportBuffer);

        FloatBuffer positionNearBuffer = BufferUtils.createFloatBuffer(3);
        FloatBuffer positionFarBuffer = BufferUtils.createFloatBuffer(3);

        GLU.gluUnProject(screenX, screenY, 0f, modelViewBuffer, projectionBuffer, viewportBuffer, positionNearBuffer);
        GLU.gluUnProject(screenX, screenY, 1f, modelViewBuffer, projectionBuffer, viewportBuffer, positionFarBuffer);

        Vector3d rayStart = new Vector3d(positionNearBuffer.get(0), positionNearBuffer.get(1), positionNearBuffer.get(2));
        Vector3d rayPosition = new Vector3d(rayStart);
        Vector3d rayDirection = new Vector3d(positionFarBuffer.get(0), positionFarBuffer.get(1), positionFarBuffer.get(2));
        rayDirection.sub(rayStart);
        rayDirection.normalize();

        for (int i = 0; i < 10000; i++) {
            int chunkX = (int) Math.round(rayPosition.getX()) >> 4;
            int chunkZ = (int) Math.round(rayPosition.getZ()) >> 4;
            ZonePlannerMapChunk zonePlannerMapChunk = ZonePlannerMapDataClient.INSTANCE.getChunk(
                field_146297_k.field_71441_e,
                new ZonePlannerMapChunkKey(
                    new ChunkPos(chunkX, chunkZ),
                    field_146297_k.field_71441_e.field_73011_w.getDimension(),
                    container.tile.getLevel()
                )
            );
            if (zonePlannerMapChunk != null) {
                BlockPos pos = new BlockPos(
                    Math.round(rayPosition.getX()) - (chunkX << 4),
                    Math.round(rayPosition.getY()),
                    Math.round(rayPosition.getZ()) - (chunkZ << 4)
                );
                MapColourData data = zonePlannerMapChunk.getData(pos.func_177958_n(), pos.func_177952_p());
                if (data != null && data.posY >= pos.func_177956_o()) {
                    found = new BlockPos(pos.func_177958_n() + (chunkX << 4), data.posY, pos.func_177952_p() + (chunkZ << 4));
                    break;
                }
            } else {
                break;
            }
            rayPosition.add(rayDirection);
        }
        return found;
    }

    @SuppressWarnings("PointlessBitwiseExpression")
    @Override
    protected void drawForegroundLayer() {
        camY += scaleSpeed;
        scaleSpeed *= 0.7F;
        int posX = (int) positionX;
        int posZ = (int) positionZ;
        int dimension = field_146297_k.field_71441_e.field_73011_w.getDimension();
        {
            ChunkPos chunkPos = new ChunkPos(posX >> 4, posZ >> 4);
            ZonePlannerMapChunk zonePlannerMapChunk = ZonePlannerMapDataClient.INSTANCE.getChunk(
                field_146297_k.field_71441_e,
                new ZonePlannerMapChunkKey(
                    chunkPos,
                    dimension,
                    container.tile.getLevel()
                )
            );
            BlockPos pos = null;
            if (zonePlannerMapChunk != null) {
                MapColourData data = zonePlannerMapChunk.getData(posX, posZ);
                if (data != null) {
                    pos = new BlockPos(posX, data.posY, posZ);
                }
            }
            if (pos != null && pos.func_177956_o() + 10 > camY) {
                camY = Math.max(camY, pos.func_177956_o() + 10);
            }
        }
        int x = field_147003_i;
        int y = field_147009_r;
        if (lastSelected != null) {
            String text = "X: " + lastSelected.func_177958_n() + " Y: " + lastSelected.func_177956_o() + " Z: " + lastSelected.func_177952_p();
            field_146289_q.func_78276_b(text, x + 130, y + 130, 0x404040);
        }
        int offsetX = 8;
        int offsetY = 9;
        int sizeX = 213;
        int sizeY = 100;
        GlStateManager.func_179094_E();
        GlStateManager.func_179128_n(GL11.GL_PROJECTION);
        GlStateManager.func_179094_E();
        GlStateManager.func_179096_D();
        ScaledResolution scaledResolution = new ScaledResolution(field_146297_k);
        int viewportX = (x + offsetX) * scaledResolution.func_78325_e();
        int viewportY = field_146297_k.field_71440_d - (sizeY + y + offsetY) * scaledResolution.func_78325_e();
        int viewportWidth = sizeX * scaledResolution.func_78325_e();
        int viewportHeight = sizeY * scaledResolution.func_78325_e();
        GL11.glEnable(GL11.GL_SCISSOR_TEST);
        GL11.glScissor(
            viewportX,
            viewportY,
            viewportWidth,
            viewportHeight
        );
        GlStateManager.func_179086_m(GL11.GL_DEPTH_BUFFER_BIT);
        GL11.glDisable(GL11.GL_SCISSOR_TEST);
        GlStateManager.func_179083_b(
            viewportX,
            viewportY,
            viewportWidth,
            viewportHeight
        );
        GlStateManager.func_179152_a(scaledResolution.func_78325_e(), scaledResolution.func_78325_e(), 1);
        GLU.gluPerspective(70.0F, (float) sizeX / sizeY, 1F, 10000.0F);
        GlStateManager.func_179128_n(GL11.GL_MODELVIEW);
        GlStateManager.func_179096_D();
        RenderHelper.func_74519_b();
        GlStateManager.func_179091_B();
        GlStateManager.func_179114_b(90, 1, 0, 0); // look down
        GlStateManager.func_179094_E();
        GlStateManager.func_179109_b(-positionX, -camY, -positionZ);
        GlStateManager.func_179084_k();
        GlStateManager.func_179118_c();
        GlStateManager.func_179090_x();
        int minScreenX = (x + offsetX) * scaledResolution.func_78325_e();
        int minScreenY = (scaledResolution.func_78328_b() - (y + offsetY)) * scaledResolution.func_78325_e();
        int maxScreenX = (x + offsetX + sizeX) * scaledResolution.func_78325_e();
        int maxScreenY = (scaledResolution.func_78328_b() - (y + offsetY + sizeY)) * scaledResolution.func_78325_e();
        int minChunkX = (posX >> 4) - 8;
        int minChunkZ = (posZ >> 4) - 8;
        int maxChunkX = (posX >> 4) + 8;
        int maxChunkZ = (posZ >> 4) + 8;
        // noinspection SuspiciousNameCombination
        List<ChunkPos> chunkPosBounds = Stream.of(
            Pair.of(minScreenX, minScreenY),
            Pair.of(minScreenX, maxScreenY),
            Pair.of(maxScreenX, minScreenY),
            Pair.of(maxScreenX, maxScreenY)
        )
            .map(p -> rayTrace(p.getLeft(), p.getRight()))
            .filter(Objects::nonNull)
            .map(ChunkPos::new)
            .collect(Collectors.toList());
        for (ChunkPos chunkPos : chunkPosBounds) {
            if (chunkPos.field_77276_a < minChunkX) {
                minChunkX = chunkPos.field_77276_a;
            }
            if (chunkPos.field_77275_b < minChunkZ) {
                minChunkZ = chunkPos.field_77275_b;
            }
            if (chunkPos.field_77276_a > maxChunkX) {
                maxChunkX = chunkPos.field_77276_a;
            }
            if (chunkPos.field_77275_b > maxChunkZ) {
                maxChunkZ = chunkPos.field_77275_b;
            }
        }
        minChunkX--;
        minChunkZ--;
        maxChunkX++;
        maxChunkZ++;
        for (int chunkX = minChunkX; chunkX <= maxChunkX; chunkX++) {
            for (int chunkZ = minChunkZ; chunkZ <= maxChunkZ; chunkZ++) {
                ZonePlannerMapRenderer.INSTANCE.getChunkGlList(
                    new ZonePlannerMapChunkKey(
                        new ChunkPos(chunkX, chunkZ),
                        dimension,
                        container.tile.getLevel()
                    )
                ).ifPresent(GlStateManager::func_179148_o);
            }
        }

        BlockPos found = null;
        int foundColor = 0;
        if (Mouse.getX() >= minScreenX &&
            Mouse.getY() <= minScreenY &&
            Mouse.getX() <= maxScreenX &&
            Mouse.getY() >= maxScreenY) {
            found = rayTrace(Mouse.getX(), Mouse.getY());
        }
        if (found != null) {
            ZonePlannerMapChunk zonePlannerMapChunk = ZonePlannerMapDataClient.INSTANCE.getChunk(
                field_146297_k.field_71441_e,
                new ZonePlannerMapChunkKey(
                    new ChunkPos(found),
                    field_146297_k.field_71441_e.field_73011_w.getDimension(),
                    container.tile.getLevel()
                )
            );
            if (zonePlannerMapChunk != null) {
                MapColourData data = zonePlannerMapChunk.getData(found.func_177958_n(), found.func_177952_p());
                if (data != null) {
                    foundColor = data.colour;
                }
            }
        }

        if (found != null) {
            GlStateManager.func_179097_i();
            GlStateManager.func_179147_l();
            GlStateManager.func_187409_d(GL11.GL_FRONT_AND_BACK, GL11.GL_LINE);
            GlStateManager.func_187441_d(2);
            int r = (int) (((foundColor >> 16) & 0xFF) * 0.7);
            int g = (int) (((foundColor >> 8) & 0xFF) * 0.7);
            int b = (int) (((foundColor >> 0) & 0xFF) * 0.7);
            int a = 0x77;
            ZonePlannerMapRenderer.INSTANCE.setColor(r << 16 | g << 8 | b << 0 | a << 24);
            BufferBuilder builder = Tessellator.func_178181_a().func_178180_c();
            builder.func_181668_a(GL11.GL_QUADS, DefaultVertexFormats.field_181706_f);
            ZonePlannerMapRenderer.INSTANCE.drawBlockCuboid(builder, found.func_177958_n(), found.func_177956_o(), found.func_177952_p());
            Tessellator.func_178181_a().func_78381_a();
            GlStateManager.func_187409_d(GL11.GL_FRONT_AND_BACK, GL11.GL_FILL);
            GlStateManager.func_179084_k();
            GlStateManager.func_179126_j();
        }

        GlStateManager.func_179140_f();
        GlStateManager.func_179147_l();

        for (int i = 0; i < container.tile.layers.length; i++) {
            if (getPaintbrushBrush() != null && getPaintbrushBrush().colour.func_176765_a() != i) {
                continue;
            }
            ZonePlan layer = container.tile.layers[i];
            if (getPaintbrushBrush() != null && getPaintbrushBrush().colour.func_176765_a() == i && bufferLayer != null) {
                layer = bufferLayer;
            }
            if (!layer.getChunkPoses().isEmpty()) {
                Tessellator.func_178181_a().func_178180_c().func_181668_a(GL11.GL_QUADS, DefaultVertexFormats.field_181706_f);
                for (int chunkX = minChunkX; chunkX <= maxChunkX; chunkX++) {
                    for (int chunkZ = minChunkZ; chunkZ <= maxChunkZ; chunkZ++) {
                        ChunkPos chunkPos = new ChunkPos(chunkX, chunkZ);
                        for (int blockX = chunkPos.func_180334_c(); blockX <= chunkPos.func_180332_e(); blockX++) {
                            for (int blockZ = chunkPos.func_180333_d(); blockZ <= chunkPos.func_180330_f(); blockZ++) {
                                if (!layer.get(
                                    blockX - container.tile.func_174877_v().func_177958_n(),
                                    blockZ - container.tile.func_174877_v().func_177952_p()
                                )) {
                                    continue;
                                }
                                int height;
                                ZonePlannerMapChunk zonePlannerMapChunk = ZonePlannerMapDataClient.INSTANCE.getChunk(
                                    field_146297_k.field_71441_e,
                                    new ZonePlannerMapChunkKey(
                                        chunkPos,
                                        dimension,
                                        container.tile.getLevel()
                                    )
                                );
                                if (zonePlannerMapChunk != null) {
                                    MapColourData data = zonePlannerMapChunk.getData(blockX, blockZ);
                                    if (data != null) {
                                        height = data.posY;
                                    } else {
                                        continue;
                                    }
                                } else {
                                    continue;
                                }
                                int color = EnumDyeColor.func_176764_b(i).func_193350_e();
                                int r = (color >> 16) & 0xFF;
                                int g = (color >> 8) & 0xFF;
                                int b = (color >> 0) & 0xFF;
                                int a = 0x55;
                                ZonePlannerMapRenderer.INSTANCE.setColor(r << 16 | g << 8 | b << 0 | a << 24);
                                ZonePlannerMapRenderer.INSTANCE.drawBlockCuboid(
                                    Tessellator.func_178181_a().func_178180_c(),
                                    blockX,
                                    height + 0.1,
                                    blockZ,
                                    height,
                                    0.6
                                );
                            }
                        }
                    }
                }
                Tessellator.func_178181_a().func_78381_a();
            }
        }
        GlStateManager.func_179084_k();
        GlStateManager.func_179140_f();
        GlStateManager.func_179098_w();

        lastSelected = found;
        GlStateManager.func_179121_F();
        GlStateManager.func_179101_C();
        GlStateManager.func_179128_n(GL11.GL_PROJECTION);
        GlStateManager.func_179083_b(0, 0, field_146297_k.field_71443_c, field_146297_k.field_71440_d);
        GlStateManager.func_179121_F();
        GlStateManager.func_179128_n(GL11.GL_MODELVIEW);
        GlStateManager.func_179121_F();
        RenderHelper.func_74518_a();
        GlStateManager.func_179084_k();
    }
}
